import { exec, spawn } from 'child_process';
import yargs from 'yargs';
import { hideBin } from 'yargs/helpers';

async function execCommand(cmd) {
    return new Promise((resolve) => {
        exec(cmd, (error, stdout, stderr) => {
            if (error) {
                console.error(`exec error: ${error}`);
                return resolve('');
            }
            if (stderr) {
                console.error(`exec stderr: ${stderr}`);
                return resolve('');
            }
            resolve(stdout.toString());
        });
    });
}
class CommandPipeChain {
    commands = [];
    getLastSpawnProcess() {
        return this.commands[this.commands.length - 1];
    }
    add(cmd, args) {
        const lastSpawnProcess = this.getLastSpawnProcess();
        const spawnProcess = spawn(cmd, args);
        if (lastSpawnProcess) {
            lastSpawnProcess.stdout.pipe(spawnProcess.stdin);
        }
        this.commands.push(spawnProcess);
        return this;
    }
    log() {
        this.commands.forEach((process) => {
            process.stdout.on(`data`, (data) => console.log(data.toString()));
        });
        return this;
    }
    get() {
        const lastSpawnProcess = this.getLastSpawnProcess();
        return new Promise((resolve, reject) => {
            let result = '';
            lastSpawnProcess.stdout.on('data', (data) => {
                result += data.toString();
            });
            lastSpawnProcess.stderr.on('data', (data) => {
                console.error(`stderr: ${data}`);
                reject(data);
            });
            lastSpawnProcess.on('close', () => resolve(result));
        });
    }
}

var SearchType;
(function (SearchType) {
    SearchType["Message"] = "message";
    SearchType["Content"] = "content";
    SearchType["All"] = "all";
})(SearchType || (SearchType = {}));
var HashType;
(function (HashType) {
    HashType["Blob"] = "blob";
    HashType["Commit"] = "commit";
    HashType["Tree"] = "tree";
})(HashType || (HashType = {}));
const PIPE_CONCURRENCY = 40;
const isWindows = process.platform === 'win32';

function isContentType(type) {
    return type === SearchType.Content;
}
function isMessageType(type) {
    return type === SearchType.Message;
}
function isAllType(type) {
    return type === SearchType.All;
}

const yargsArgv = yargs(hideBin(process.argv))
    .option('s', {
    alias: 'search',
    demand: false,
    describe: '请输入你的查找内容（-s可省略）',
    type: 'string',
})
    .option('t', {
    choices: [SearchType.Content, SearchType.Message, SearchType.All],
    alias: 'type',
    default: SearchType.All,
    describe: '要匹配的查找类型，message匹配git的提交信息，content匹配文件的更改内容，all会同时查找两者',
    type: 'string',
})
    .option('l', {
    alias: 'length',
    default: 80,
    describe: '返回匹配字符串的内容长度值',
    type: 'number',
    skipValidation: false
})
    .usage('Usage: ggc -s <search-content> [-t <type>] [-l <length>]')
    .example('ggc "hello world"', '将查找所有内容更改或提交信息里包含‘hello world’的git-hash，每条对象最多返回80个字符长度的内容')
    .example('ggc -s "hello world"', '结果同上')
    .example('ggc -s "hello world" -t message', '将查找所有提交信息包含‘hello world’的git-hash，每条对象最多返回80个字符长度的内容')
    .example('ggc -s "hello world" -t content -l 50', '将查找所有内容更改包含‘hello world’的git-hash，每条对象最多返回50个字符长度的内容')
    .help('h')
    .alias('h', 'help')
    .epilog('copyright 2024')
    .argv;
const { _, search: search$1, type, length } = yargsArgv;
const argv = {
    search: search$1 ?? _[0] ?? '',
    contentLength: Number.isNaN(length) ? 80 : length,
    searchType: type,
};

const GREP = isWindows ? 'findstr' : 'grep';
const { search, contentLength, searchType } = argv;
const isSearchMessage = isAllType(searchType) || isMessageType(searchType);
const isSearchContent = isAllType(searchType) || isContentType(searchType);
async function handleGrep() {
    console.time('exec');
    await Promise.all([
        grepLostFound(),
        isSearchMessage && grepMessage(),
        isSearchContent && grepContent()
    ]);
    console.timeEnd('exec');
}
async function getLostFoundhashs() {
    const allHashs = (await execCommand(`git fsck --lost-found`)).trim().split('\n').map(line => line.split(/ +/g));
    const blobHashs = [];
    const commitHashs = [];
    allHashs.forEach(([_, hashType, hash]) => {
        (hashType === HashType.Blob) && (blobHashs.push(hash));
        (hashType === HashType.Commit) && (commitHashs.push(hash));
    });
    return {
        blobHashs,
        commitHashs,
    };
}
async function grepLostFound() {
    const { blobHashs, commitHashs } = await getLostFoundhashs();
    const [blobHashMap, commitHashMap,] = await Promise.all([
        getHashMap(blobHashs, HashType.Blob),
        getHashMap(commitHashs, HashType.Commit),
    ]);
    console.log(`\n--lost-found--`);
    if (blobHashMap) {
        console.log(`\nblob-hash:\n`);
        console.table(blobHashMap);
    }
    console.log(`\ncommit-hash:\n`);
    console.table(commitHashMap);
}
async function getHashMap(hashs, hashType) {
    const hashMap = {};
    if (isMessageType(searchType) && hashType === HashType.Blob) {
        return null;
    }
    const showParametersMap = {
        [SearchType.All]: [],
        [SearchType.Message]: ['-s', '--format=%s'],
        [SearchType.Content]: ['--format=""'],
    };
    for (let i = 0; i * PIPE_CONCURRENCY < hashs.length; i++) {
        const hashFragment = hashs.slice(PIPE_CONCURRENCY * i, PIPE_CONCURRENCY * (i + 1));
        await Promise.all(hashFragment.map(async (hash) => {
            const matchStr = await new CommandPipeChain()
                .add('git', ['show', hash, ...showParametersMap[searchType]])
                .add(GREP, [search])
                .get();
            if (matchStr) {
                hashMap[hash] = matchStr.slice(0, contentLength);
            }
        }));
    }
    return hashMap;
}
async function grepContent() {
    const res = (await execCommand(`git log --all --oneline -S ${JSON.stringify(search)}`)).trim();
    console.log(`\n--commit-content--\n`, res);
}
async function grepMessage() {
    const res = (await execCommand(`git log --all --oneline --grep ${JSON.stringify(search)}`)).trim();
    console.log(`\n--commit-message--\n`, res);
}
if (search) {
    handleGrep();
}
else {
    console.error('请输入: ggc -s <查找内容>');
}
